from __future__ import annotations

import builtins
import typing
from abc import ABC, abstractmethod
from pathlib import PurePosixPath
from typing import Any

if typing.TYPE_CHECKING:
    from _hydrogenlib_resource_system.core.system import CoreResourceSystem

__all__ = [
    'ResourceProvider',
    'Resource',
]


class ResourceProvider(ABC):
    @abstractmethod
    def list(self, path: PurePosixPath, query: dict[str, Any],
             resource_system: CoreResourceSystem) -> builtins.list: ...

    @abstractmethod
    def get(self, path: PurePosixPath, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> Resource | None: ...

    @abstractmethod
    def set(self, path: PurePosixPath, data: Any, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> None: ...

    @abstractmethod
    def exists(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool: ...

    @abstractmethod
    def remove(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem): ...

    def close(self):
        pass

    def __del__(self):
        self.close()


class Resource(ABC):
    url: str | Any = None

    @abstractmethod
    def __fspath__(self) -> str:
        ...

    def open(
            self,
            mode='r',
            encoding=None,
            buffering=-1,
            errors: str | None = None,
            opener: typing.Callable[[str, int], int] | None = None) -> typing.IO:
        """
        打开资源内容

        :return: 资源 IO 流，实现了 typing.IO 接口
        """
        return open(self, mode, buffering, encoding, errors, opener=opener)

    def release(self) -> None:
        """
        释放资源

        有些资源可能限制只能有一个访问者
        调用此函数可以主动释放资源，但这可能会让你无法继续正常使用资源

        当对象被回收时，此函数自动被调用
        """
        ...

    def parse_as[T](self, type: type[T] | typing.Callable[[typing.Self], T]) -> T:
        """
        获取资源，并尝试转换为 type 类型

        :param type: 一个可调用对象，调用时接受一个类型为 Resource 的参数，返回一个转换后的对象
        :return: 经过转换的对象
        :raises: 不定
        """
        if hasattr(type, '__from_resource__'):
            return type.__from_resource__(type)
        else:
            return type(self)

    @property
    def text(self) -> str:
        """
        读取自身的 text 内容
        """
        with self.open() as f:
            return f.read()

    @property
    def binary(self) -> builtins.bytes:
        """
        读取自身的 bytes 内容
        """
        with self.open('rb') as f:
            return f.read()

    @property
    def size(self) -> int:
        raise NotImplemented

    def __bytes__(self):
        return self.binary

    def __str__(self):
        return self.text

    def __repr__(self):
        return f"""
<Resource: {self.url}>
size: 
content: {self.text}
content(bytes): {self.binary}
"""



from . import core
from .builtin_providers import URLProvider


class TypedResourceSystem[T]:
    __slots__ = ('_system', '_type')

    def __init__(self, system, type):
        self._system = system
        self._type = type

    def get(self, url, **query) -> T:
        return self._system.get(url, **query).parse_as(self._type)


class ResourceSystem:
    def __init__(self):
        self._system = core.system.CoreResourceSystem()

    def get(self, url, **query):
        return self._system.get(url, **query)

    def remove(self, url, **kwargs):
        return self._system.remove(url, **kwargs)

    def set(self, url, data, **query):
        return self._system.set(url, data, **query)

    def list(self, url, **query):
        return self._system.list(url, **query)

    def exists(self, url, **query):
        return self._system.exists(url, **query)

    def mount(self, url, provider=URLProvider):
        return self._system.mount(url, provider)

    def bind(self, dst, src):
        return self._system.mount(
            dst, URLProvider(src)
        )

    def open(self, url, mode='r', encoding='utf-8', **query):
        return self.get(url, **query).open(mode, encoding)

    def __getitem__[T](self, item: type[T]) -> TypedResourceSystem[T]:
        return TypedResourceSystem(self, item)

    def close(self):
        self._system.close()



from __future__ import annotations

import builtins
import os
import tempfile
import typing
from pathlib import PurePosixPath, Path
from typing import Any

from .core.provider import Resource, ResourceProvider

if typing.TYPE_CHECKING:
    from _hydrogenlib_resource_system.core.system import CoreResourceSystem

import zipfile


class LocalResource(Resource):
    def __init__(self, local_path: str | Path):
        self.local_path = local_path

    def __fspath__(self) -> str:
        return str(self.local_path)

    @property
    def size(self) -> int:
        return os.stat(self).st_size


class URLProvider(ResourceProvider):
    def __init__(self, prefix):
        self.prefix = PurePosixPath(prefix)

    def fullpath(self, path):
        return self.prefix / path

    def get(self, path: PurePosixPath, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> Resource | None:
        return resource_system.get(
            self.fullpath(path)
        )

    def list(self, path: PurePosixPath, query: dict[str, Any],
             resource_system: CoreResourceSystem) -> builtins.list:
        return resource_system.list(
            self.fullpath(path)
        )

    def set(self, path: PurePosixPath, data: Any, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> None:
        resource_system.set(
            self.fullpath(path), data
        )

    def exists(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool:
        return resource_system.exists(
            self.fullpath(path)
        )

    def remove(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool:
        return resource_system.remove(
            self.fullpath(path)
        )


class FSProvider(ResourceProvider):
    # 拼接路径
    def __init__(self, root: str | Path):
        self.root = Path(root)

    def fullpath(self, path):
        path = self.root / str(path)[1:]  # Fix for windows: PurePosixPath 的根目录是 /，但是这样会导致拼接的时候被识别成盘符根目录
        # 比如 'C:/xxx/xxx/xx' + '/resource' 会变成 'C:/resource'

        if not path.is_relative_to(self.root):  # Fix: 防止出现相对路径攻击
            raise ValueError(f'{path} is not in {self.root}')

        return path

    def list(self, path: PurePosixPath, query: dict[str, Any],
             resource_system: CoreResourceSystem) -> builtins.list:
        return list(self.fullpath(path).iterdir())

    def get(self, path: PurePosixPath, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> Resource | None:
        return LocalResource(
            self.fullpath(path)
        )

    def set(self, path: PurePosixPath, data: Any, query: dict[str, Any],
            resource_system: CoreResourceSystem) -> None:
        if isinstance(data, str):
            fmode = 'w'
        elif isinstance(data, bytes | memoryview | bytearray):
            fmode = 'wb'
        else:
            raise TypeError(f'unwritable data type: {type(data)!r}')
        f = open(self.fullpath(path), fmode)
        f.write(data)
        f.close()

    def exists(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool:
        return self.fullpath(path).exists()

    def remove(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem):
        self.fullpath(path).unlink()


class AppProvider(FSProvider):
    def __init__(self, app_dir=None):
        super().__init__(app_dir or Path.cwd())


class TempProvider(ResourceProvider):
    def __init__(self, suffix=None, prefix=None):
        self.temp_dir_manager = tempfile.TemporaryDirectory(suffix, prefix)
        self.temp_dir = Path(self.temp_dir_manager.name)

    def list(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> builtins.list:
        return []

    def get(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> Resource | None:
        return LocalResource(
            self.temp_dir / path
        )

    def set(self, path: PurePosixPath, data: Any, query: dict[str, Any], resource_system: CoreResourceSystem) -> None:
        raise NotImplementedError('Use the `get` function instead')

    def exists(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool:
        return (
                self.temp_dir / path
        ).exists()

    def remove(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem):
        (self.temp_dir / path).unlink()

    def close(self):
        if self.temp_dir_manager:
            self.temp_dir_manager.cleanup()
            self.temp_dir_manager = None

    def __del__(self):
        self.close()


class ZipResource(Resource):
    def __init__(self, zipio):
        self._io = zipio

    def release(self) -> None:
        self._io.close()

    def __fspath__(self):
        raise NotImplemented


class ZipProvider(ResourceProvider):
    def list(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> builtins.list:
        raise NotImplemented

    def get(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> Resource | None:
        pwd = query.get('pwd', None)
        return ZipResource(
            self._zip.open(str(path), force_zip64=True, pwd=pwd)
        )

    def set(self, path: PurePosixPath, data: Any, query: dict[str, Any], resource_system: CoreResourceSystem) -> None:
        raise NotImplemented

    def exists(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem) -> bool:
        raise NotImplemented

    def remove(self, path: PurePosixPath, query: dict[str, Any], resource_system: CoreResourceSystem):
        raise NotImplemented

    def __init__(self, file):
        self._zipfile = file
        self._zip = zipfile.ZipFile(file)

    def close(self):
        self._zip.close()



import dataclasses
from pathlib import PurePosixPath


@dataclasses.dataclass
class URLInfo:
    scheme: str
    path: PurePosixPath


def parse_url(url: str) -> URLInfo:
    sep_index = url.find(":")
    if sep_index == -1:
        scheme = url
        path = '/'
    else:
        scheme = url[:sep_index]
        path = url[sep_index + 1:]

    return URLInfo(
        scheme,
        PurePosixPath(path)
    )



from pathlib import PurePosixPath
from typing import NamedTuple

from . import errors
from .provider import ResourceProvider
from .url import parse_url, URLInfo


class MountInfo(NamedTuple):
    path: PurePosixPath
    provider: ResourceProvider

    def get(self, path, query, rs):
        return self.provider.get(path, query, rs)

    def remove(self, path, query, rs):
        return self.provider.remove(path, query, rs)

    def list(self, path, query, rs):
        return self.provider.list(path, query, rs)

    def set(self, path, data, query, rs):
        return self.provider.set(path, data, query, rs)

    def exists(self, path, query, rs):
        return self.provider.exists(path, query, rs)


class CoreResourceSystem:

    def __init__(self):
        self._mounttab = {}  # type: dict[str, list[MountInfo]]

    def mount(self, prefix: str, provider: ResourceProvider | type[ResourceProvider]):
        URLInfo = parse_url(prefix)
        scheme, path = URLInfo.scheme, URLInfo.path

        if scheme not in self._mounttab:
            self._mounttab[scheme] = []

        if isinstance(provider, type):
            provider = provider()

        self._mounttab[scheme].append(MountInfo(path, provider))

    def find_mount(self, url: str) -> tuple[MountInfo | None, URLInfo]:
        url_info = parse_url(url)
        scheme, path = url_info.scheme, url_info.path

        if scheme in self._mounttab:
            for mountinfo in self._mounttab[scheme]:
                mpath = mountinfo.path
                if path.is_relative_to(mpath):
                    return mountinfo, url_info

        return None, url_info

    def get_mount(self, url: str) -> tuple[MountInfo, URLInfo]:
        minfo, url_info = self.find_mount(url)
        if minfo is None:
            raise errors.ResourceNotFound(url)
        return minfo, url_info

    def get(self, url: str, **query):
        minfo, url_info = self.get_mount(url)
        return minfo.get(url_info.path, query, self)

    def remove(self, url: str, **query):
        minfo, url_info = self.get_mount(url)
        return minfo.remove(url_info.path, query, self)

    def set(self, url: str, data, **query):
        minfo, url_info = self.get_mount(url)
        return minfo.set(url_info.path, data, query, self)

    def list(self, url: str, **query):
        minfo, url_info = self.get_mount(url)
        return minfo.list(url_info.path, query, self)

    def exists(self, url: str, **query):
        minfo, url_info = self.get_mount(url)
        return minfo.exists(url_info.path, query, self)

    def close(self):
        for ls in self._mounttab.values():
            for minfo in ls:
                provider = minfo.provider
                provider.close()



from __future__ import annotations

from typing import Callable, TypedDict

from .core.provider import Resource, ResourceProvider


def registry_resource_type[T](type_: type[T]):
    """
    注册资源类型装饰器
    类型被注册后, 可以通过 resource_system.get("...").parse_as(T) 的格式来直接获取这个类型的实例
    :param type_: 需要注册的类型
    :return: 一个装饰器, 应通过此装饰器来指定 ``Resource`` -> ``Type`` 的转换函数
    """

    def decorator(func: Callable[[type[T], Resource], T]):
        type_.__from_resource__ = func
        return func

    return decorator


class _WrappedResourceType:
    def __init__(self, typ, convert_func):
        self._typ = typ
        self._cf = convert_func

    def __getattribute__(self, item):
        if item in {'_typ', '_cf', '__call__'}:
            return super().__getattribute__(item)

        return getattr(self._typ, item)

    def __call__(self, *args, **kwargs):
        return self._cf(*args, **kwargs)


def wrap_type[T](typ: type[T], convert_func) -> type[T]:
    return _WrappedResourceType(typ, convert_func)


# class _MTB(TypedDict):
#     provider: ResourceProvider | type[ResourceProvider]
#     children: MountTab
#
#
# type MountTab = dict[str, str | MountTab | ResourceProvider | type[ResourceProvider] | _MTB]
#
#
# def make_system_by_tab(
#         tab: MountTab
# ):
#     from .system import ResourceSystem
#     system = ResourceSystem()
#
#
#
#     return system
